package org.bitbucket.bradleysmithllc.etlunit.feature.database;

import org.bitbucket.bradleysmithllc.etlunit.RuntimeSupport;
import org.bitbucket.bradleysmithllc.etlunit.TestExecutionError;
import org.bitbucket.bradleysmithllc.etlunit.feature.file.FileRuntimeSupport;
import org.bitbucket.bradleysmithllc.etlunit.feature.file.RequestedFmlNotFoundException;
import org.bitbucket.bradleysmithllc.etlunit.io.FileBuilder;
import org.bitbucket.bradleysmithllc.etlunit.io.file.DataFileManager;
import org.bitbucket.bradleysmithllc.etlunit.io.file.DataFileSchema;
import org.bitbucket.bradleysmithllc.etlunit.metadata.MetaDataArtifact;
import org.bitbucket.bradleysmithllc.etlunit.metadata.MetaDataContext;
import org.bitbucket.bradleysmithllc.etlunit.metadata.MetaDataManager;
import org.bitbucket.bradleysmithllc.etlunit.metadata.MetaDataPackageContext;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestValueObject;

import javax.inject.Inject;
import java.io.File;
import java.io.IOException;

public class DatabaseRuntimeSupportImpl implements DatabaseRuntimeSupport
{
	private MetaDataManager metaDataManager;
	private MetaDataContext dataMetaContext;
	private MetaDataContext scriptsMetaContext;

	private RuntimeSupport runtimeSupport;
	private FileRuntimeSupport fileRuntimeSupport;
	private DataFileManager dataFileManager;

	@Inject
	public void receiveMetaDataManager(MetaDataManager manager)
	{
		metaDataManager = manager;
		dataMetaContext = metaDataManager.getMetaData().getOrCreateContext("data");
		scriptsMetaContext= metaDataManager.getMetaData().getOrCreateContext("database");
	}

	@Inject
	public void setRuntimeSupport(RuntimeSupport runtimeSupport)
	{
		this.runtimeSupport = runtimeSupport;
	}

	@Inject
	public void receiveFileRuntimeSupport(FileRuntimeSupport manager)
	{
		fileRuntimeSupport = manager;
	}

	@Inject
	public void receiveDataFileManager(DataFileManager manager)
	{
		dataFileManager = manager;
	}

	@Override
	public File getSourceDataDirectoryForCurrentTest()
	{
		return getSourceDataDirectory(runtimeSupport.getCurrentlyProcessingTestPackage());
	}

	@Override
	public File getSourceScriptsDirectoryForCurrentTest()
	{
		return getSourceScriptsDirectory(runtimeSupport.getCurrentlyProcessingTestPackage());
	}

	@Override
	public File getSourceDataSetForCurrentTest(String dataSetName, String use)
	{
		return getSourceDataSet(runtimeSupport.getCurrentlyProcessingTestPackage(), dataSetName, use);
	}

	@Override
	public File getSourceScriptForCurrentTest(String scriptName, String use)
	{
		return getSourceScript(runtimeSupport.getCurrentlyProcessingTestPackage(), scriptName, use);
	}

	public File getSourceDataDirectory(String package_)
	{
		return new FileBuilder(runtimeSupport.getTestResourceDirectory(package_, "data")).mkdirs().file();
	}

	public File getSourceScriptsDirectory(String package_)
	{
		return new FileBuilder(runtimeSupport.getTestResourceDirectory(package_, "sql")).mkdirs().file();
	}

	public File getSourceDataSet(String package_, String dataSetName, String use)
	{
		File delim = new FileBuilder(getSourceDataDirectory(package_)).name(dataSetName + ".delimited").file();

		if (delim != null && runtimeSupport.isTestActive())
		{
			MetaDataPackageContext dataPackageContext = dataMetaContext.createPackageContextForCurrentTest(MetaDataPackageContext.path_type.test_source);

			try {
				MetaDataArtifact art = dataPackageContext.createArtifact("data", getSourceDataDirectoryForCurrentTest());
				art.createContent(delim.getName()).referencedByCurrentTest(use);
			} catch (IOException e) {
				throw new IllegalStateException("Error getting data context", e);
			}
		}

		return delim;
	}

	public File getSourceScript(String package_, String scriptName, String use)
	{
		File sql = new FileBuilder(getSourceScriptsDirectory(package_)).name(scriptName).file();

		if (sql != null && runtimeSupport.isTestActive())
		{
			MetaDataPackageContext dataPackageContext = dataMetaContext.createPackageContextForCurrentTest(MetaDataPackageContext.path_type.test_source);
			MetaDataArtifact art;

			try {
				art = dataPackageContext.createArtifact("sql", getSourceDataDirectoryForCurrentTest());
				art.createContent(scriptName).referencedByCurrentTest(use);
			} catch (IOException e) {
				throw new IllegalStateException("Error getting data context", e);
			}
		}

		return sql;
	}

	@Override
	public File getCachedMetaData(DatabaseConnection conn) {
		return new FileBuilder(getGeneratedSourceDirectory(conn, null)).name(conn.getId() + "_metadata.json").file();
	}

	public File getGeneratedSourceDirectory(DatabaseConnection conn, String mode)
	{
		File dir = new FileBuilder(runtimeSupport.getGeneratedSourceDirectory("database")).subdir(conn.getId()).subdir(mode == null ? "[default]" : mode).mkdirs().file();

		return dir;
	}

	public File getGeneratedDataSet(DatabaseConnection conn, String mode, String dataSetName)
	{
		File dir = new FileBuilder(getGeneratedSourceDirectory(conn, mode)).name(dataSetName + ".delimited").file();

		return dir;
	}

	public DataFileSchema createRelationalMetaInfo(DatabaseConnection dbConn, String schema, String name)
	{
		return createRelationalMetaInfo(dbConn.getId(), schema, name);
	}

	private String createSchemaKey(String connId, String schema, String name)
	{
		return connId + "-" + schema + "-" + name;
	}

	public DataFileSchema createRelationalMetaInfo(String connId, String schema, String name)
	{
		return dataFileManager.createDataFileSchema(createSchemaKey(connId, schema, name));
	}

	public DataFileSchema getRelationalMetaInfo(DatabaseConnection dbConn, String schema, String name, String masterName, ETLTestValueObject etlObj) throws RequestedFmlNotFoundException {
		return getRelationalMetaInfo(dbConn.getId(), schema, name, masterName, etlObj);
	}

	public DataFileSchema getRelationalMetaInfo(String connId, String schema, String name, String masterName, ETLTestValueObject etlObj) throws RequestedFmlNotFoundException {
		return fileRuntimeSupport.locateReferenceFileSchema(masterName, createSchemaKey(connId, schema, name), etlObj);
	}

	public void persistDataFileSchema(DataFileSchema schema) throws IOException
	{
		fileRuntimeSupport.persistGeneratedDataFileSchema(schema);
	}
}