package org.bitbucket.bradleysmithllc.etlunit.feature.database.db;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.google.gson.stream.JsonWriter;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.DatabaseImplementation;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.RelationalColumn;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.RelationalDataSet;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.extract.ExtractRequest;
import org.bitbucket.bradleysmithllc.etlunit.io.file.DataFileSchema;

import java.io.IOException;
import java.util.*;

public class TableImpl implements Table
{
	private final Schema schema;
	private final Catalog catalog;
	private transient DataFileSchema dataFileSchema;

	private String name;
	private type type;

	private String sqlScriptSource;

	private final List<TableColumn> tableColumns = new ArrayList<TableColumn>();
	private final Map<String, TableColumn> columnNames = new HashMap<String, TableColumn>();

	private final List<TableColumn> primaryKeyColumns = new ArrayList<TableColumn>();
	private final List<TableColumn> orderKeyColumns = new ArrayList<TableColumn>();

	public TableImpl(Catalog catalog, Schema schema, String name, Table.type type) {
		this.schema = schema;
		this.catalog = catalog;
		this.name = name;
		this.type = type;
	}

	public TableImpl(Catalog catalog, Schema schema) {
		this(catalog, schema, null, null);
	}

	@Override
	public Schema getSchema() {
		return schema;
	}

	@Override
	public Catalog getCatalog() {
		return catalog;
	}

	@Override
	public String getName() {
		return name;
	}

	@Override
	public type getType() {
		return type;
	}

	@Override
	public String getQualifiedName()
	{
		return getSchema().getName() + "." + getName();
	}

	@Override
	public String getDBLocalName() {
		StringBuilder stb = new StringBuilder();

		if (catalog.getName() != null)
		{
			stb.append(catalog.getName());
			stb.append('.');
		}

		if (schema.getName() != null)
		{
			stb.append(schema.getName());
			stb.append('.');
		}

		stb.append(name);

		return stb.toString();
	}

	@Override
	public List<TableColumn> getTableColumns() {
		return Collections.unmodifiableList(tableColumns);
	}

	@Override
	public List<String> getTableColumnNames() {
		List<String> names = new ArrayList<String>();

		for (TableColumn tc : tableColumns)
		{
			names.add(tc.getName());
		}

		return names;
	}

	@Override
	public void resetPrimaryKeyColumns() {
		primaryKeyColumns.clear();
	}

	public String toString()
	{
		return (schema == null ? "" : (schema.toString() + ".")) + name;
	}

	@Override
	public void toJson(JsonWriter writer) throws IOException {
		writer.name("name").value(name);
		writer.name("type").value(type.name());
		writer.name("sql-script").value(sqlScriptSource);

		writer.name("columns");
		writer.beginArray();

		for (TableColumn column : tableColumns)
		{
			writer.beginObject();
			column.toJson(writer);
			writer.endObject();
		}

		writer.endArray();

		writer.name("primary-key-columns");
		writer.beginArray();

		for (TableColumn column : primaryKeyColumns)
		{
			writer.value(column.getName());
		}

		writer.endArray();
	}

	@Override
	public void fromJson(JsonNode node) {
		name = node.get("name").asText();
		type = Table.type.valueOf(node.get("type").asText());

		JsonNode jsonNode = node.get("sql-script");

		sqlScriptSource = (jsonNode == null || jsonNode.isNull()) ? null : jsonNode.asText();

		ArrayNode anode = (ArrayNode) node.get("columns");

		Iterator<JsonNode> it = anode.iterator();

		while (it.hasNext())
		{
			JsonNode onode = it.next();

			TableColumnImpl timpl = new TableColumnImpl(this);

			timpl.fromJson(onode);

			addColumn(timpl);
		}

		anode = (ArrayNode) node.get("primary-key-columns");

		it = anode.iterator();

		while (it.hasNext())
		{
			JsonNode onode = it.next();

			addPrimaryKeyColumn(getColumn(onode.asText()));
		}
	}

	@Override
	public void addColumn(TableColumn timpl) {
		if (columnNames.containsKey(timpl.getName()))
		{
			throw new IllegalArgumentException("Column named [" + timpl.getName() + "] already present");
		}

		tableColumns.add(timpl);
		columnNames.put(timpl.getName(), timpl);
	}

	@Override
	public boolean hasColumn(String columnName) {
		return columnNames.containsKey(columnName);
	}

	@Override
	public TableColumn getColumn(String columnName) {
		if (columnName == null)
		{
			throw new NullPointerException("Column name is null");
		}

		if (!hasColumn(columnName))
		{
			throw new IllegalArgumentException("Column [" + columnName + "] not in column list");
		}

		return columnNames.get(columnName);
	}

	@Override
	public void addPrimaryKeyColumn(TableColumn column) {
		if (column == null)
		{
			throw new NullPointerException("Column is null");
		}

		if (primaryKeyColumns.contains(column))
		{
			throw new IllegalArgumentException("Column [" + column.getName() + "] already added to primary keys");
		}

		if (!tableColumns.contains(column))
		{
			throw new IllegalArgumentException("Column [" + column.getName() + "] not in column list");
		}

		primaryKeyColumns.add(column);
		touch();
	}

	@Override
	public List<TableColumn> getPrimaryKeyColumns() {
		return Collections.unmodifiableList(primaryKeyColumns);
	}

	@Override
	public void resetOrderColumns() {
		orderKeyColumns.clear();
		touch();
	}

	private void touch() {
		if (catalog != null)
		{
			catalog.getDatabase().touch();
		}
	}

	@Override
	public void addOrderKeyColumn(TableColumn column) {
		if (column == null)
		{
			throw new NullPointerException("Column is null");
		}

		if (orderKeyColumns.contains(column))
		{
			throw new IllegalArgumentException("Column [" + column.getName() + "] already added to order keys");
		}

		if (!tableColumns.contains(column))
		{
			throw new IllegalArgumentException("Column [" + column.getName() + "] not in column list");
		}

		orderKeyColumns.add(column);

		touch();
	}

	@Override
	public List<TableColumn> getOrderKeyColumns() {
		return Collections.unmodifiableList(orderKeyColumns);
	}

	@Override
	public void setSqlScriptSource(String sqlScriptSource) {
		if (this.sqlScriptSource == null || !this.sqlScriptSource.equals(sqlScriptSource))
		{
			if (type != Table.type.sql)
			{
				type = Table.type.sql;
			}

			this.sqlScriptSource = sqlScriptSource;

			columnNames.clear();
			primaryKeyColumns.clear();
			tableColumns.clear();
			orderKeyColumns.clear();
			touch();
		}
	}

	public String getSqlScriptSource() {
		return sqlScriptSource;
	}

	@Override
	public DataFileSchema getDataFileSchema() {
		if (dataFileSchema == null)
		{
			throw new IllegalStateException("No schema");
		}

		return dataFileSchema;
	}

	@Override
	public void setDataFileSchema(DataFileSchema dataFileSchema) {
		this.dataFileSchema = dataFileSchema;

		if (this.dataFileSchema == null || !this.dataFileSchema.equals(dataFileSchema))
		{
			this.dataFileSchema = dataFileSchema;

			touch();
		}
	}

	@Override
	public boolean hasIdentityColumns()
	{
		for (TableColumn column : tableColumns)
		{
			if (column.isAutoIncrement())
			{
				return true;
			}
		}

		return false;
	}

	public String createSQLInsert()
	{
		return createSQLInsert(null);
	}

	@Override
	public String createSQLInsert(DatabaseImplementation dbi)
	{
		if (type == Table.type.sql)
		{
			throw new UnsupportedOperationException("Cannot insert into a synthetic data set");
		}

		StringBuffer buffer = new StringBuffer("INSERT INTO\n");
		buffer.append('\t');

		if (!schema.isVirtual() && schema.getName() != null)
		{
			buffer.append(dbi == null ? schema.getName() : dbi.escapeIdentifier(schema.getName()));
			buffer.append('.');
		}

		buffer.append(dbi == null ? name : dbi.escapeIdentifier(name));
		buffer.append("\n(\n");

		for (int i = 0; i < tableColumns.size(); i++)
		{
			buffer.append('\t');
			buffer.append(dbi == null ? tableColumns.get(i).getName() : dbi.escapeIdentifier(tableColumns.get(i).getName()));

			if (i != (tableColumns.size() - 1))
			{
				buffer.append(',');
			}

			buffer.append('\n');
		}

		buffer.append(")\n");

		buffer.append("VALUES\n(\n\t");

		for (int i = 0; i < tableColumns.size(); i++)
		{
			if (i != 0)
			{
				buffer.append(',');
			}

			buffer.append('?');
		}
		buffer.append("\n)");

		return buffer.toString();
	}

	@Override
	public String createSQLSelect(DatabaseImplementation dbi)
	{
		return createSQLSelect(dbi, true);
	}

	@Override
	public String createSQLSelect()
	{
		return createSQLSelect(null, true);
	}

	@Override
	public String createSQLSelect(boolean ordered)
	{
		return createSQLSelect(null, ordered);
	}

	@Override
	public String createSQLSelect(DatabaseImplementation dbi, boolean ordered)
	{
		List<TableColumn> actualList = new ArrayList<TableColumn>(tableColumns);
		List<TableColumn> actualOrderBy = new ArrayList<TableColumn>(primaryKeyColumns);

		if (orderKeyColumns.size() != 0)
		{
			actualOrderBy.clear();
			actualOrderBy.addAll(orderKeyColumns);
		}
		else if (actualOrderBy.size() == 0)
		{
			actualOrderBy.addAll(actualList);
		}

		if (type != Table.type.sql)
		{
			StringBuffer buffer = new StringBuffer("SELECT\n");

			for (int i = 0; i < actualList.size(); i++)
			{
				buffer.append('\t');
				buffer.append(dbi == null ? actualList.get(i).getName() : dbi.escapeIdentifier(actualList.get(i).getName()));

				if (i != (actualList.size() - 1))
				{
					buffer.append(',');
				}

				buffer.append('\n');
			}

			buffer.append("FROM\n");
			buffer.append('\t');

			if (schema.getName() != null && !schema.isVirtual())
			{
				buffer.append(dbi == null ? schema.getName() : dbi.escapeIdentifier(schema.getName()));
				buffer.append('.');
			}

			buffer.append(dbi == null ? name : dbi.escapeIdentifier(name));

			if (ordered)
			{
				buffer.append("\nORDER BY\n");

				for (int i = 0; i < actualOrderBy.size(); i++)
				{
					buffer.append('\t');
					buffer.append(dbi == null ? actualOrderBy.get(i).getName() : dbi.escapeIdentifier(actualOrderBy.get(i).getName()));

					if (i != (actualOrderBy.size() - 1))
					{
						buffer.append(',');
					}

					buffer.append('\n');
				}
			}

			return buffer.toString();
		}

		return sqlScriptSource;
	}

	@Override
	public Table createSubView(List<String> list, ExtractRequest.ColumnListMode listMode)
	{
		if (type == Table.type.synthetic)
		{
			throw new UnsupportedOperationException("Can't create view of sql data set");
		}

		List<TableColumn> actualList = new ArrayList<TableColumn>(tableColumns);
		List<TableColumn> actualKeys = new ArrayList<TableColumn>(primaryKeyColumns);
		List<TableColumn> actualOrders = new ArrayList<TableColumn>(orderKeyColumns);
		Map<String, TableColumn> actualNames = new HashMap<String, TableColumn>(columnNames);

		if (actualKeys.size() == 0)
		{
			actualKeys.addAll(actualList);
		}

		if (list != null)
		{
			if (listMode == ExtractRequest.ColumnListMode.INCLUDE)
			{
				actualList.clear();
				actualKeys.clear();
				actualOrders.clear();
				actualNames.clear();
			}

			for (String col : list)
			{
				TableColumn column = columnNames.get(col);

				if (column == null)
				{
					throw new IllegalArgumentException("Column not exported: " + col);
				}

				switch (listMode)
				{
					case INCLUDE:
						// check the column
						actualList.add(column);
						actualNames.put(column.getName(), column);

						// check the primary key
						if (primaryKeyColumns.contains(column))
						{
							actualKeys.add(column);
						}

						// check the order columns
						if (orderKeyColumns.contains(column))
						{
							actualOrders.add(column);
						}
						break;
					case EXCLUDE:
						// check the column
						actualList.remove(column);
						actualNames.remove(column.getName());

						// check the primary key
						if (actualKeys.contains(column))
						{
							actualKeys.remove(column);
						}

						// check the order columns
						if (orderKeyColumns.contains(column))
						{
							actualOrders.remove(column);
						}
						break;
					default:
						throw new RuntimeException("Refactor error");
				}
			}
		}

		TableImpl timpl = new TableImpl(catalog, schema, name, Table.type.synthetic);

		timpl.tableColumns.addAll(actualList);
		timpl.primaryKeyColumns.addAll(actualKeys);
		timpl.orderKeyColumns.addAll(actualOrders);
		timpl.columnNames.putAll(actualNames);

		return timpl;
	}
}